@@ -1,40 +1,151 @@
 from collections import defaultdict
 from typing import Any, List
+import re
+import yaml
 
 from docutils import nodes
 from docutils.nodes import Node
-from docutils.parsers.rst import directives
-from rdflib import Graph, Literal, Namespace  # type: ignore
+from docutils.parsers.rst import directives  # type: ignore
+from rdflib import Graph, URIRef, Literal, BNode, Namespace  # type: ignore
 from rdflib.namespace import RDF, RDFS  # type: ignore
 from sphinx import addnodes
 from sphinx.directives import ObjectDescription
+from sphinx.directives.code import CodeBlock
 from sphinx.domains import Domain, Index
 from sphinx.roles import XRefRole
 from sphinx.util.docfields import DocFieldTransformer
 from sphinx.util.docutils import SphinxDirective
 from sphinx.util.nodes import make_refnode
+from sphinx.util import logging
+
+logger = logging.getLogger(__name__)
 
 
-class TTL(SphinxDirective):
+def parse_composed_of(value):
+    """Parse composed_of option."""
+    if value is None:
+        return []
+    return [x.strip() for x in value.split()]
+
+
+def parse_consumes_or_produces(value):
+    """Parse consumes and products options."""
+    if value is None:
+        return []
+
+    if "\n" in value or value.strip().startswith("{"):
+        # Complex definitions, one per line -- or a single YAML-style definition
+        items = [_parse_item(x.strip()) for x in value.strip().split("\n")]
+    else:
+        # A space-separated list of basic names
+        items = [
+            {"object": x.strip()}
+            for x in value.split()
+        ]
+
+    return items
+
+
+ITEM_STRING_REGEX = re.compile(r"""
+    ^\s*
+    ([^= \t]+)               # -> Object type
+    \s*
+    (?:                      # Optional amount section
+        =
+        \s*
+        ([0-9.eE-]+)         # -> Amount
+        (?:                  # Optional unit section
+            \s*
+            ([^{]+?)         # -> Unit
+            \s*
+        )?
+    )?
+    ({.+})?                  # Optional YAML section
+    $
+""", re.VERBOSE)
+
+
+def _parse_item(item):
+    if isinstance(item, str):
+        match = ITEM_STRING_REGEX.match(item)
+        if match:
+            extra = {}
+            if match.group(4):
+                try:
+                    extra = yaml.safe_load(match.group(4))
+                except yaml.YAMLError:
+                    pass
+            return {
+                "object": match.group(1),
+                "amount": float(match.group(2)) if match.group(2) else None,
+                "unit": match.group(3),
+                **extra,
+            }
+        else:
+            # Try parsing whole thing as yaml dict
+            try:
+                d = yaml.safe_load(item)
+                if not isinstance(d, dict):
+                    raise ValueError("YAML data should be dictionary")
+                return d
+            except yaml.YAMLError:
+                pass
+
+            return item
+    elif isinstance(item, list):
+        return {
+            "object": item[0],
+            "amount": item[1],
+            "unit": item[2],
+        }
+    elif isinstance(item, dict):
+        return item
+    else:
+        raise ValueError("cannot parse item: %r" % item)
+
+
+def expand_consumes_produces_amounts(defs, *items):
+    """Expand Python expressions in cleaned-up options."""
+    defs_ns = {}
+    exec(defs, defs_ns)
+
+    # Expand amounts in produces/consumes lists using the defs
+    result = [
+        [eval_amount(x, defs_ns) for x in item_list]
+        for item_list in items
+    ]
+
+    return result
+
+
+def eval_amount(item, namespace):
+    """Evaluate expressions within the "amount" field of the item.
+
+    WARNING: not safe for use with untrusted input!
+    """
+    if isinstance(item, dict) and isinstance(item.get("amount"), str):
+        amount = eval(item["amount"], {}, dict(namespace))
+        return {**item, "amount": amount}
+    return item
+
+
+class TTL(CodeBlock):
     has_content = True
 
     def run(self):
-        if not hasattr(self.env, "probs_all_ttl"):
-            self.env.probs_all_ttl = []
+        g = get_or_setup_graph(self.env, self.config)
+        preamble = "\n".join('@prefix %s: <%s> .\n' % (prefix, uri)
+                             for prefix, uri in g.namespaces())
+        input_data = preamble + "\n" + "\n".join(self.content)
+        try:
+            g.parse(data=input_data, format="text/turtle")
+        except SyntaxError as err:
+            logger.warning("Cannot parse TTL block: %s", err,
+                           location=(self.env.docname, self.lineno))
 
-        self.env.probs_all_ttl.append(
-            {
-                "docname": self.env.docname,
-                "lineno": self.lineno,
-                "content": self.content,
-                # 'target': targetnode,
-            }
-        )
-        # print(self.env.docname, self.lineno)
-        # print(self.content)
-
-        paragraph_node = nodes.literal_block(text="\n".join(self.content))
-        return [paragraph_node]
+        # Force language
+        self.arguments = ["turtle"]
+        return super().run()
 
 
 class StartSubProcessesDirective(SphinxDirective):
@@ -81,22 +192,29 @@ class EndSubProcessesDirective(SphinxDirective):
             self.env.probs_parent = []
         if self.env.probs_parent:
             leaving = self.env.probs_parent.pop()
-            paragraph_node = nodes.literal_block(
-                text=f'Ending sub processes of "{leaving}"'
+            _, _, leaving_id = leaving.rpartition("/")
+            nesting_depth = len(self.env.probs_parent)
+            paragraph_node = nodes.paragraph(
+                text=f'Ending sub processes of "{leaving_id}"',
+                classes=["system", "end-sub-processes", f"nested-{nesting_depth}"],
             )
         else:
             paragraph_node = nodes.literal_block(text="Nothing to end!")
         return [paragraph_node]
 
 
+
 class EndSubObjectsDirective(SphinxDirective):
     def run(self):
         if not hasattr(self.env, "probs_parent_object"):
             self.env.probs_parent_object = []
         if self.env.probs_parent_object:
             leaving = self.env.probs_parent_object.pop()
-            paragraph_node = nodes.literal_block(
-                text=f'Ending sub objects of "{leaving}"'
+            _, _, leaving_id = leaving.rpartition("/")
+            nesting_depth = len(self.env.probs_parent_object)
+            paragraph_node = nodes.paragraph(
+                text=f'Ending sub objects of "{leaving_id}"',
+                classes=["system", "end-sub-objects", f"nested-{nesting_depth}"],
             )
         else:
             paragraph_node = nodes.literal_block(text="Nothing to end!")
@@ -104,6 +222,8 @@ class EndSubObjectsDirective(SphinxDirective):
 
 
 PROBS = Namespace("https://ukfires.org/probs/ontology/")
+PROBS_RECIPE = Namespace("https://ukfires.org/probs/ontology/recipe/")
+QUANTITYKIND = Namespace("http://qudt.org/vocab/quantitykind/")
 
 
 def remove_ns(ns, uri):
@@ -111,6 +231,22 @@ def remove_ns(ns, uri):
         return uri[len(ns) :]
 
 
+def get_or_setup_graph(env, config):
+    """Return the rdflib Graph or set up a new one as needed."""
+    if env is None or not hasattr(env, "probs_graph"):
+        g = Graph()
+        g.bind("sys", Namespace(config.probs_rdf_system_prefix))
+        g.bind("probs", PROBS)
+        g.bind("rec", PROBS_RECIPE)
+        for prefix, uri in config.probs_rdf_extra_prefixes.items():
+            g.bind(prefix, uri)
+        if env is not None:
+            env.probs_graph = g
+    else:
+        g = env.probs_graph
+    return g
+
+
 class SystemObjectDescription(ObjectDescription):
     has_content = True
     required_arguments = 1
@@ -217,6 +353,17 @@ class SystemObjectDescription(ObjectDescription):
             )
         return sig
 
+    def _system_id_link(self, sys_id, within=None):
+        """Insert a cross reference to another object/process."""
+        refnode = addnodes.pending_xref('', refdomain="system", refexplicit=False,
+                                        reftype="ref", reftarget=sys_id)
+        refnode += nodes.inline(sys_id, sys_id)
+        if within is not None:
+            wrapper = within("", "")
+            wrapper += refnode
+            return wrapper
+        return refnode
+
 
 class Process(SystemObjectDescription):
     # has_content = True
@@ -224,22 +371,41 @@ class Process(SystemObjectDescription):
     option_spec = {
         "label": directives.unchanged_required,
         "become_parent": directives.flag,
-        "consumes": directives.unchanged,
-        "produces": directives.unchanged,
-        "composed_of": directives.unchanged,
+        "consumes": parse_consumes_or_produces,
+        "produces": parse_consumes_or_produces,
+        "composed_of": parse_composed_of,
+        "defs": directives.unchanged,
     }
     signature_prefix = "Process: "
 
     def transform_content(self, contentnode):
-        if "consumes" in self.options:
-            text = f"Consumes: {self.options['consumes']}"
-            contentnode += nodes.paragraph(text, text)
-        if "produces" in self.options:
-            text = f"Produces: {self.options['produces']}"
-            contentnode += nodes.paragraph(text, text)
+        # XXX Should refactor this to reduce duplication and/or read from the
+        # graph
+        defs = self.options.get("defs", "")
+        consumes, produces = expand_consumes_produces_amounts(
+            defs, self.options.get("consumes", []), self.options.get("produces", []))
+
+        if consumes:
+            contentnode += nodes.paragraph("Consumes: ", "Consumes: ")
+            contentnode += self._recipe_table(consumes)
+        if produces:
+            contentnode += nodes.paragraph("Produces: ", "Produces: ")
+            contentnode += self._recipe_table(produces)
         if hasattr(self.env, "probs_parent") and self.env.probs_parent:
-            text = f"Parent: {self.env.probs_parent[-1]}"
-            contentnode += nodes.paragraph(text, text)
+            _, _, parent_id = self.env.probs_parent[-1].rpartition("/")
+            p = nodes.paragraph("", "Parent: ")
+            p += self._system_id_link(parent_id)
+            contentnode += p
+
+    def _recipe_table(self, objects):
+        header_rows = [[nodes.literal("", "Object"), nodes.literal("", "Amount")]]
+        table_data = [
+            [self._system_id_link(obj["object"], nodes.paragraph),
+             nodes.literal("", "%.1f %s" % (obj["amount"], obj["unit"]))
+             if "amount" in obj else ""]
+            for obj in objects
+        ]
+        return build_table_from_list(header_rows + table_data, header_rows=1)
 
     def get_nesting_depth(self):
         if hasattr(self.env, "probs_parent"):
@@ -254,18 +420,12 @@ class Process(SystemObjectDescription):
             recipes = self.env.get_domain("system")
             recipes.add_process(
                 sig,
-                self.options.get("consumes", "").split(),
-                self.options.get("produces", "").split(),
+                self.options.get("consumes", []),
+                self.options.get("produces", []),
             )
 
     def define_graph(self, uri_str):
-        if not hasattr(self.env, "probs_graph"):
-            g = Graph()
-            g.bind("sys", self.SYS)
-            g.bind("probs", PROBS)
-            self.env.probs_graph = g
-        else:
-            g = self.env.probs_graph
+        g = get_or_setup_graph(self.env, self.config)
 
         if not hasattr(self.env, "probs_parent"):
             self.env.probs_parent = []
@@ -283,7 +443,7 @@ class Process(SystemObjectDescription):
         # ComposedOf relationships
         if self.env.probs_parent:
             g.add((self.env.probs_parent[-1], PROBS.processComposedOf, uri))
-        for child in self.options.get("composed_of", "").split():
+        for child in self.options.get("composed_of", []):
             if child.startswith("*"):
                 # include children of the named process -- this is expanded
                 # later as a postprocessing step once all processes are defined.
@@ -293,19 +453,69 @@ class Process(SystemObjectDescription):
                 child_uri = self.SYS[child]
                 g.add((uri, PROBS.processComposedOf, child_uri))
 
-        for obj in self.options.get("consumes", "").split():
-            obj_uri = getattr(self.SYS, obj)
-            g.add((uri, PROBS.consumes, obj_uri))
+        # Recipes (inputs and outputs)
+        # First expand any expressions
+        defs = self.options.get("defs", "")
+        consumes, produces = expand_consumes_produces_amounts(
+            defs, self.options.get("consumes", []), self.options.get("produces", []))
 
-        for obj in self.options.get("produces", "").split():
-            obj_uri = getattr(self.SYS, obj)
-            g.add((uri, PROBS.produces, obj_uri))
+        recipe_consumes, recipe_produces = [], []
+        _process_inputs_outputs(g, self.SYS, uri, "consumes", consumes, recipe_consumes)
+        _process_inputs_outputs(g, self.SYS, uri, "produces", produces, recipe_produces)
+        if recipe_consumes or recipe_produces:
+            recipe = BNode()
+            g.add((uri, PROBS_RECIPE.hasRecipe, recipe))
+            for item in recipe_consumes:
+                g.add((recipe, PROBS.consumes, item))
+            for item in recipe_produces:
+                g.add((recipe, PROBS.produces, item))
 
         if "become_parent" in self.options:
             # print("xxx become parent")
             self.env.probs_parent.append(uri)
 
 
+def _process_inputs_outputs(g, SYS, uri, relation, objects, recipe_items):
+    for obj in objects:
+        obj_uri = getattr(SYS, obj["object"])
+        g.add((uri, PROBS[relation], obj_uri))
+
+        if "amount" in obj:
+            # Have a recipe
+
+            # XXX only support kg for now
+            if "unit" in obj and obj["unit"] != "kg":
+                logger.error("Unsupported unit %r for object %r in recipe for %r",
+                             obj["unit"], obj["object"], uri)
+
+            item = BNode()
+            g.add((item, PROBS_RECIPE.object, obj_uri))
+            g.add((item, PROBS_RECIPE.quantity, Literal(obj["amount"])))
+            g.add((item, PROBS_RECIPE.metric, QUANTITYKIND.Mass))
+            recipe_items.append(item)
+
+
+def parse_traded(value):
+    """Check the value of the :traded: option is valid."""
+    if value is None:
+        return (False, False)
+    value = value.lower()
+    imp = value.startswith("import")
+    exp = value.startswith("export")
+    if (value in ("both", "yes", "true") or
+        "import" in value and "export" in value):
+        imp = exp = True
+    return (imp, exp)
+
+
+def parse_equivalent(value):
+    """Convert list of uris."""
+    if value is None:
+        value = ""
+    items = [x.strip() for x in value.split()]
+    return items
+
+
 class Object(SystemObjectDescription):
     has_content = True
     required_arguments = 1
@@ -313,6 +523,8 @@ class Object(SystemObjectDescription):
         "label": directives.unchanged,
         "become_parent": directives.flag,
         "parent_object": directives.unchanged,
+        "traded": parse_traded,
+        "equivalent": parse_equivalent,
     }
     signature_prefix = "Object: "
 
@@ -326,9 +538,10 @@ class Object(SystemObjectDescription):
 
     def transform_content(self, contentnode):
         if hasattr(self.env, "probs_parent_object") and self.env.probs_parent_object:
-            text = f"Parent: {remove_ns(self.SYS, self.env.probs_parent_object[-1])}"
-            ref = nodes.reference()
-            contentnode += nodes.paragraph(text, text)
+            _, _, parent_id = self.env.probs_parent_object[-1].rpartition("/")
+            p = nodes.paragraph("", "Parent: ")
+            p += self._system_id_link(parent_id)
+            contentnode += p
 
     def get_nesting_depth(self):
         if hasattr(self.env, "probs_parent_object"):
@@ -344,13 +557,7 @@ class Object(SystemObjectDescription):
             recipes.add_object(sig, [])
 
     def define_graph(self, uri_str):
-        if not hasattr(self.env, "probs_graph"):
-            g = Graph()
-            g.bind("sys", self.SYS)
-            g.bind("probs", PROBS)
-            self.env.probs_graph = g
-        else:
-            g = self.env.probs_graph
+        g = get_or_setup_graph(self.env, self.config)
 
         if not hasattr(self.env, "probs_parent_object"):
             self.env.probs_parent_object = []
@@ -358,6 +565,7 @@ class Object(SystemObjectDescription):
         uri = getattr(self.SYS, uri_str)
 
         g.add((uri, RDF.type, PROBS.Object))
+        g.add((uri, RDF.type, PROBS.ReferenceObject))
         if "label" in self.options:
             g.add((uri, RDFS.label, Literal(self.options["label"])))
             g.add((uri, PROBS.objectName, Literal(self.options["label"])))
@@ -375,6 +583,41 @@ class Object(SystemObjectDescription):
             # print("xxx become parent object")
             self.env.probs_parent_object.append(uri)
 
+        if "traded" in self.options:
+            imp, exp = self.options["traded"]
+            if imp != exp:
+                logger.error("Currently objects must be either fully traded"
+                             "(imports and exports) or not at all")
+            g.add((uri, PROBS.objectIsTraded, Literal(imp or exp)))
+
+        if "equivalent" in self.options:
+            for item in self.options["equivalent"]:
+                item_uri = self.parse_uri(item)
+                g.add((uri, PROBS.objectEquivalentTo, item_uri))
+
+
+    def parse_uri(self, item):
+        """Convert a string to a URIRef.
+
+        A blank prefix or bare id refers to the namespace given by the
+        `probs_rdf_system_prefix` config variable.
+
+        A missing suffix means the same as the object currently being defined.
+
+        """
+        if item and item[0] == "<" and item[-1] == ">":
+            return URIRef(item[1:-1])
+        prefix, _, item_id = item.rpartition(":")
+        if not prefix:
+            ns = Namespace(self.config.probs_rdf_system_prefix)
+        else:
+            ns = Namespace(self.config.probs_rdf_extra_prefixes[prefix])
+        if not item_id:
+            signatures = self.get_signatures()
+            assert len(signatures) == 1, "only assuming 1 signature can be given"
+            item_id = signatures[0]
+        return getattr(ns, item_id)
+
 
 class ObjectIndex(Index):
     """Index of objects."""
@@ -398,21 +641,28 @@ class ObjectIndex(Index):
         process_produces = self.domain.data["process_produces"]
         object_processes = defaultdict(list)
 
-        # flip from recipe_ingredients to ingredient_recipes
+        # Add the objects initially; this is necessary for any objects which are
+        # NOT linked to processes to be included.
+        for obj in objects.values():
+            dispname = obj[0]
+            object_processes[dispname] = []
+
+        # Add in all the places that an object is consumed or produced by a
+        # process
         for process_name, objs in process_consumes.items():
             for obj in objs:
-                if ("object." + obj) not in objects:
+                if ("object." + obj["object"]) not in objects:
                     msg = "WARNING: Object {} consumed by process {} is not defined"
-                    print(msg.format(obj, process_name))
+                    print(msg.format(obj["object"], process_name))
                     continue
-                object_processes[obj].append((process_name, "consumed"))
+                object_processes[obj["object"]].append((process_name, "consumed"))
         for process_name, objs in process_produces.items():
             for obj in objs:
-                if ("object." + obj) not in objects:
+                if ("object." + obj["object"]) not in objects:
                     msg = "WARNING: Object {} produced by process {} is not defined"
-                    print(msg.format(obj, process_name))
+                    print(msg.format(obj["object"], process_name))
                     continue
-                object_processes[obj].append((process_name, "produced"))
+                object_processes[obj["object"]].append((process_name, "produced"))
 
         # convert the mapping of objects to processes to produce the expected
         # output, shown below, using the object name as a key to group
@@ -500,9 +750,10 @@ class SystemDomain(Domain):
             todocname = match[0][0]
             targ = match[0][1]
 
+            # print("system domain: found %s %s %r xref" % (todocname, targ, target))
             return make_refnode(builder, fromdocname, todocname, targ, contnode, targ)
         else:
-            print("Awww, found nothing")
+            print("system domain: found nothing for %r xref" % target)
             return None
 
     def add_process(self, signature, consumes, produces):
@@ -528,3 +779,56 @@ class SystemDomain(Domain):
         self.data["objects"].append(
             (name, signature, "Object", self.env.docname, anchor, 0)
         )
+
+
+# Adapted from docutils ListTable directive
+def build_table_from_list(table_data,
+                          # col_widths,
+                          header_rows, stub_columns=0, widths="auto"):
+    """
+    :param table_data: list of lists giving table data
+    :param header_rows: list of header rows
+    :param stub_columns: number of columns to mark as "stubs"
+    """
+
+    table = nodes.table()
+
+    max_cols = len(table_data[0])
+    col_widths = [100 // max_cols] * max_cols
+    # if widths == 'auto':
+    #     table['classes'] += ['colwidths-auto']
+    # elif widths: # "grid" or list of integers
+    #     table['classes'] += ['colwidths-given']
+    table['classes'] += ['colwidths-auto']
+
+    tgroup = nodes.tgroup(cols=max_cols)
+    table += tgroup
+
+    for col_width in col_widths:
+        colspec = nodes.colspec()
+        # if col_width is not None:
+        #     colspec.attributes['colwidth'] = col_width
+        if stub_columns:
+            colspec.attributes['stub'] = 1
+            stub_columns -= 1
+        tgroup += colspec
+
+    rows = []
+    for row in table_data:
+        row_node = nodes.row()
+        for cell in row:
+            entry = nodes.entry()
+            entry += cell
+            row_node += entry
+        rows.append(row_node)
+
+    if header_rows:
+        thead = nodes.thead()
+        thead.extend(rows[:header_rows])
+        tgroup += thead
+
+    tbody = nodes.tbody()
+    tbody.extend(rows[header_rows:])
+    tgroup += tbody
+
+    return table
